unit BigText;
{ TBigText 1.1  (c) 1995 by Gerry Skolnik (skolnik@kapsch.co.at)
                    Portions (c) 1995 by Danny Thorpe

  This is a simple component to display up to 32767 lines of text. Each line
  has its own dedicated foreground and background color and can be 255 chars
  long. Theoretically this amounts to about 8MB of data and beats the TMemo's
  measly 32kB, however, no editing functions are available.

  TBigList is a no-frills TList mutant. I've implemented most of the
  essential functions. Before fine-tuning I'd like to wait for Windows 95 /
  Delphi 95, just in case TBigList is made redundant then.

  The limitation of TBigText is caused by the Windows API scrolling functions
  insisting on being passed integer values, thus reducing the maximum amount
  of lines a scrollbar can handle to 32767. However, display problems start
  as soon as line 32750. As I couldn't see much difference between 32750 and
  32767 lines, I haven't bothered to track this down. Be my guest.

  TBigText is FreeWare. You may use it freely at your own risk in any
  kind of environment. This component is not to be sold at any charge, and
  must be distributed along with the source code.

  The scrolling routines were taken from Danny Thorpe's TConsole object.

  BTW: while I claim the copyright to the original source code, this does
  not mean that you may not modify or enhance it. Just add your credits,
  and if you think you came up with some major improvement that the Delphi
  community might find useful, upload it at some Delphi site.
  Of course, any enhancement/modification must be released as Freeware.

  property MaxLines
           if set to 0, as much lines as memory permits are included. The
           absolute maximum, however, is 32767. If set to something else,
           TBigText will limit itself to that many lines.

  property PurgeLines
           determines how to handle the situation when no more lines can be
           added (line count reached Maxlines value or we ran out of memory).
           if set to 0, an exception is raised. If set to something different
           (default 200) the number of lines specified by PurgeLines are
           deleted, the TBigList objects are packed, and most likely more
           lines can be added (though the first ones will be lost).
           This option is useful for logging windows.

  property Count
           run-time read-only. If the Lines and StringColor counts are equal, this
           property holds the number of lines in TBigText. If the two counts are
           unequal, there's something wrong and the property holds a value of -1.

  procedure AddLine(LineString: string; FCol, BCol: TColor; UpdateDisplay: boolean);
           The essential routine to insert lines into TBigText.
           LineString   : the text to be inserted
           FCol         : forground color
           BCol         : background color
           UpdateDisplay: if true, TBigText will scroll to the last line
                          (where the new line will be added), and update
                          its display. This is not recommended if lots of
                          lines are to be included in a loop.

  procedure LoadFromFile(FileName: TFileName);
           Loads a file into TBigText. Every line will have the default colors
           clWindowText, clWindow.

  procedure Print
           prints all lines on the specified printer. Haven't checked this out, though.

  procedure ChangeColor(Index: longint; OldFCol, OldBCol, NewFCol, NewBCol: TColor);
           changes the colors of the line at Index, but only if the current colors
           match OldFCol and OldBCol (FCol = foreground color, BCol = background color).

  the following procedures do pretty much the same as the accodring TList methods:

  procedure Clear;
  procedure Delete(Index: longint);
  procedure Remove(Index: longint);
  procedure Pack;
}
interface

uses WinTypes, WinProcs, Messages, Classes, Controls, Printers,
     Forms, Graphics, SysUtils;

type
  {$M+}
  TStringColor = class
  public
    FColor : TColor;
    BColor : TColor;
  end;
  TBigList = class
  private
    function GetCapacity: longint;
    function GetCount: longint;
    function GetItems(Index: longint): pointer;
    procedure SetItems(Index: longint; const Item: pointer);
  protected
    ListCount : integer;
    TheLines  : array[0..3] of TList;
  published
    property Capacity: longint read GetCapacity;
    property Count: longint read GetCount;
  public
    property Items[Index: longint]: pointer read GetItems write SetItems;
    constructor Create;
    destructor Destroy;
    class function ClassName: string;
    function Add(Item: Pointer): longint;
    procedure Delete(Index: longint);
    procedure Remove(Index: longint);
    procedure Pack;
    procedure Clear;
    function First: pointer;
    function Last: pointer;
  end;
  {$M-}
  TBigText = class(TCustomControl)
  private
    FFont: TFont;
    FMaxLines: word;
    FPurgeLines: word;
    FColor : TColor;
    procedure DoScroll(Which, Action, Thumb: Integer);
    procedure WMHScroll(var M: TWMHScroll); message wm_HScroll;
    procedure WMVScroll(var M: TWMVScroll); message wm_VScroll;
    procedure WMSize(var M: TWMSize); message wm_Size;
    procedure WMGetDlgCode(var M: TWMGetDlgCode); message wm_GetDlgCode;
    procedure SetFont(F: TFont);
    function GetCount: longint;
  protected
    FRange: TPoint;
    FOrigin: TPoint;
    FClientSize: TPoint;
    FCharSize: TPoint;
    FOverhang: Integer;
    FPageSize: integer;
    Lines: TBigList;
    StringColor: TBigList;
    procedure Paint; override;
    procedure SetScrollbars;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
		        X, Y: Integer); override;
  published
    procedure RecalcRange;
    procedure FontChanged(Sender: TObject);
    property Font: TFont read FFont write SetFont;
    property Align;
    property ParentColor;
    property MaxLines: word read FMaxLines write FMaxLines default 0;
    property PurgeLines: word read FPurgeLines write FPurgeLines default 200;
    property Color: TColor read FColor write FColor default clWindow;
    property Count: longint read GetCount;
  public
    constructor Create(AnOwner: TComponent); override;
    destructor Destroy; override;
    procedure ScrollTo(X, Y: Integer);
    procedure AddLine(LineString: string; FCol, BCol: TColor; UpdateDisplay: boolean);
    procedure Delete(Index: longint);
    procedure Clear;
    procedure Print;
    procedure LoadFromFile(FileName: TFileName);
    function GetLine(Index: longint): string;
    procedure ChangeColor(Index: longint; OldFCol, OldBCol, NewFCol, NewBCol: TColor);
  end;

procedure Register;

implementation

{ Scroll key definition record }

type
  TScrollKey = record
    sKey: Byte;
    Ctrl: Boolean;
    SBar: Byte;
    Action: Byte;
  end;

{ Scroll keys table }

const
  ScrollKeyCount = 12;
  ScrollKeys: array[1..ScrollKeyCount] of TScrollKey = (
    (sKey: vk_Left;  Ctrl: False; SBar: sb_Horz; Action: sb_LineUp),
    (sKey: vk_Right; Ctrl: False; SBar: sb_Horz; Action: sb_LineDown),
    (sKey: vk_Left;  Ctrl: True;  SBar: sb_Horz; Action: sb_PageUp),
    (sKey: vk_Right; Ctrl: True;  SBar: sb_Horz; Action: sb_PageDown),
    (sKey: vk_Home;  Ctrl: False; SBar: sb_Horz; Action: sb_Top),
    (sKey: vk_End;   Ctrl: False; SBar: sb_Horz; Action: sb_Bottom),
    (sKey: vk_Up;    Ctrl: False; SBar: sb_Vert; Action: sb_LineUp),
    (sKey: vk_Down;  Ctrl: False; SBar: sb_Vert; Action: sb_LineDown),
    (sKey: vk_Prior; Ctrl: False; SBar: sb_Vert; Action: sb_PageUp),
    (sKey: vk_Next;  Ctrl: False; SBar: sb_Vert; Action: sb_PageDown),
    (sKey: vk_Home;  Ctrl: True;  SBar: sb_Vert; Action: sb_Top),
    (sKey: vk_End;   Ctrl: True;  SBar: sb_Vert; Action: sb_Bottom));

function Min(X, Y: Integer): Integer;
begin
  if X < Y then Min := X else Min := Y;
end;

function Max(X, Y: Integer): Integer;
begin
  if X > Y then Max := X else Max := Y;
end;

{<<<<<<<<<<<<<<<<<<<< TBigList >>>>>>>>>>>>>>>>>>>>>>>}

constructor TBigList.Create;
begin
  ListCount := 0;
  TheLines[ListCount] := TList.Create;
end;

destructor TBigList.Destroy;
var
  i: integer;
begin
  for i := 0 to ListCount do
    TheLines[i].Free;
end;

class function TBigList.ClassName: string;
begin
  ClassName := 'TBigList';
end;

function TBigList.GetCapacity: longint;
var
  i: integer;
  j: longint;
begin
  j := 0;
  for i := 0 to ListCount do
    inc(j, TheLines[i].Capacity);
  GetCapacity := j;
end;

function TBigList.GetCount: longint;
var
  i: integer;
  j: longint;
begin
  j := 0;
  for i := 0 to ListCount do
    inc(j, TheLines[i].Count);
  GetCount := j;
end;

function TBigList.Add(Item: Pointer): longint;
var
  i: integer;
  j: longint;
begin
  try
    TheLines[ListCount].Add(Item);
    j := 0;
    for i := 0 to ListCount do
      inc(j, TheLines[ListCount].Count);
    Add := j - 1;
  except
    try
      inc(ListCount);
      TheLines[ListCount] := TList.Create;
      TheLines[ListCount].Add(Item);
      j := 0;
      for i := 0 to ListCount do
        inc(j, TheLines[i].Count);
      Add := j - 1;
    except
      j := 0;
      for i := 0 to (ListCount - 1) do
        inc(j, TheLines[i].Count);
      raise EOutOfResources.Create('Out of Memory at line ' + IntToStr(j));
      Add := -1;
    end;
  end;
end;

procedure TBigList.Delete(Index: longint);
var
  i: integer;
begin
  if Index > Count then
    raise ERangeError.Create('TBigList Index out of bounds')
  else
  begin
    i := 0;
    while Index > (TheLines[i].Count - 1) do
    begin
      dec(Index, TheLines[i].Count);
      inc(i);
    end;
    TheLines[i].Delete(Index);
  end;
end;

procedure TBigList.Remove(Index: longint);
begin
  Delete(Index);
end;

procedure TBigList.Pack;
var
  i       : integer;
  j       : longint;
  ListFull: boolean;
begin
  TheLines[0].Pack;
  i := 0;
  while (i < ListCount) do
  begin
    try
      TheLines[i].Add(TheLines[i + 1].Items[0]);
      TheLines[i + 1].Delete(0);
    except
      inc(i);
    end;
  end;
  TheLines[i].Pack;
  for i := ListCount downto 1 do
  begin
    if TheLines[i].Count = 0 then
      TheLines[i].Free;
  end;
end;

procedure TBigList.Clear;
var
  i: integer;
begin
  for i := 1 to ListCount do
    TheLines[ListCount].Free;
  ListCount := 0;
  TheLines[ListCount].Clear;
end;

function TBigList.First: pointer;
begin
  First := TheLines[0].Items[0];
end;

function TBigList.Last: pointer;
begin
  Last := TheLines[ListCount].Items[TheLines[ListCount].Count - 1];
end;

function TBigList.GetItems(Index: longint): pointer;
var
  i: integer;
begin
  if Index > Count then
    raise ERangeError.Create('TBigList Index out of bounds')
  else
  begin
    i := 0;
    while Index > (TheLines[i].Count - 1) do
    begin
      dec(Index, TheLines[i].Count);
      inc(i);
    end;
    GetItems := TheLines[i].Items[Index];
  end;
end;

procedure TBigList.SetItems(Index: longint; const Item: pointer);
var
  i: integer;
begin
  if Index > Count then
    raise ERangeError.Create('TBigList Index out of bounds')
  else
  begin
    i := 0;
    while Index > (TheLines[i].Count - 1) do
    begin
      dec(Index, TheLines[i].Count);
      inc(i);
    end;
    TheLines[i].Items[Index] := Item;
  end;
end;

{<<<<<<<<<<<<<<<<<<<< TBigText >>>>>>>>>>>>>>>>>>>>>>>}

constructor TBigText.Create(AnOwner: TComponent);
begin
  inherited Create(AnOwner);
  Width := 320;
  Height := 200;
  ParentColor := False;
  FFont := TFont.Create;
  FFont.Name := 'Courier';
  FFont.OnChange := FontChanged;
  FColor := clWindow;
  FMaxLines := 0;
  FPurgeLines := 200;
  FOrigin.X := 0;
  FOrigin.Y := 0;
  FontChanged(nil);
  Enabled := True;
  Lines := TBigList.Create;
  StringColor := TBigList.Create;
end;

destructor TBigText.Destroy;
begin
  Lines.Free;
  StringColor.Free;
  FFont.Free;
  inherited Destroy;
end;

procedure TBigText.FontChanged(Sender: TObject);
var
  DC: HDC;
  Save: THandle;
  Metrics: TTextMetric;
  Temp: String;
begin
  DC := GetDC(0);
  Save := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, Save);
  ReleaseDC(0, DC);
  with Metrics do
  begin
    FCharSize.X := tmAveCharWidth;
    FCharSize.Y := tmHeight + tmExternalLeading;
    FOverhang   := Max(tmOverhang, tmMaxCharWidth - tmAveCharWidth);
    RecalcRange;
    Invalidate;
  end;
end;

procedure TBigText.RecalcRange;
begin
  if HandleAllocated then
  begin
    FClientSize.X := ClientWidth div FCharSize.X;
    FClientSize.Y := ClientHeight div FCharSize.Y;
    FPageSize := FClientSize.Y;
    FRange.X := Max(0, 255 - FClientSize.X);
    FRange.Y := Max(0, Lines.Count - FClientSize.Y);
    ScrollTo(Min(FOrigin.X, FRange.X), Min(FOrigin.Y, FRange.Y));
    SetScrollBars;
  end;
end;

procedure TBigText.SetScrollBars;
begin
  if HandleAllocated then
  begin
    SetScrollRange(Handle, sb_Horz, 0, Max(1, FRange.X), False);
    SetScrollPos(Handle, sb_Horz, FOrigin.X, True);
    SetScrollRange(Handle, sb_Vert, 0, Max(1, FRange.Y), False);
    SetScrollPos(Handle, sb_Vert, FOrigin.Y, True);
  end;
end;

procedure TBigText.Paint;
var
  i: longint;
  R: TRect;
begin
  SetViewportOrg(Canvas.Handle, -FOrigin.X * FCharSize.X, 0);
  i := FOrigin.Y;
  while (i < Lines.Count) and (i < ((FOrigin.Y + FPageSize) + 1)) do
  begin
    Canvas.Font := FFont;
    Canvas.Font.Color := TStringColor(StringColor.Items[i]).FColor;
    Canvas.Brush.Color := TStringColor(StringColor.Items[i]).BColor;
    TextOut(Canvas.Handle, 0, FCharSize.Y * (i - FOrigin.Y),
            Lines.Items[i], StrLen(Lines.Items[i]));
    inc(i);
  end;
end;

procedure TBigText.DoScroll(Which, Action, Thumb: Integer);
var
  X, Y: Integer;
function GetNewPos(Pos, Page, Range: Integer): Integer;
begin
  case Action of
    sb_LineUp: GetNewPos := Pos - 1;
    sb_LineDown: GetNewPos := Pos + 1;
    sb_PageUp: GetNewPos := Pos - Page;
    sb_PageDown: GetNewPos := Pos + Page;
    sb_Top: GetNewPos := 0;
    sb_Bottom: GetNewPos := Range;
    sb_ThumbPosition,
    sb_ThumbTrack    : GetNewPos := Thumb;
  else
    GetNewPos := Pos;
  end;
end;
begin
  X := FOrigin.X;
  Y := FOrigin.Y;
  case Which of
    sb_Horz: X := GetNewPos(X, FClientSize.X div 2, FRange.X);
    sb_Vert: Y := GetNewPos(Y, FClientSize.Y, FRange.Y);
  end;
  ScrollTo(X, Y);
end;

procedure TBigText.WMHScroll(var M: TWMHScroll);
begin
  DoScroll(sb_Horz, M.ScrollCode, M.Pos);
end;

procedure TBigText.WMVScroll(var M: TWMVScroll);
begin
  DoScroll(sb_Vert, M.ScrollCode, M.Pos);
end;

procedure TBigText.WMSize(var M: TWMSize);
begin
  inherited;
  RecalcRange;
end;

procedure TBigText.ScrollTo(X, Y: Integer);
var
  R: TRect;
  OldOrigin: TPoint;
begin
  X := Max(0, Min(X, FRange.X));  { check boundaries }
  Y := Max(0, Min(Y, FRange.Y));
  if (X <> FOrigin.X) or (Y <> FOrigin.Y) then
  begin
    OldOrigin := FOrigin;
    FOrigin.X := X;
    FOrigin.Y := Y;
    if HandleAllocated then
    begin
      R := ClientRect;
      ScrollWindowEx(Handle, (OldOrigin.X - X) * FCharSize.X, (OldOrigin.Y - Y) * FCharSize.Y,
                     nil, @R, 0, @R, 0);
      if Y <> OldOrigin.Y then
        SetScrollPos(Handle, sb_Vert, Y, True);
      if X <> OldOrigin.X then
        SetScrollPos(Handle, sb_Horz, X, True);
      InvalidateRect(Handle, @R, true);
      Update;
    end;
  end;
end;

procedure TBigText.AddLine(LineString: string; FCol, BCol: TColor; UpdateDisplay: boolean);
var
  DumChar: array[0..255] of char;
  WhereY : integer;
  i      : integer;
  LeCol  : TStringColor;
begin
  if FMaxLines <> 0 then
  begin
    if (Lines.Count >= FMaxLines) or (Lines.Count > 32000) then
    begin
      if PurgeLines <> 0 then
      begin
        for i := 1 to PurgeLines do
        begin
          Lines.Delete(0);
          StringColor.Delete(0);
        end;
        Lines.Pack;
        StringColor.Pack;
      end
      else
        raise ERangeError.Create('Maximum line count at line ' + IntToStr(Lines.Count))
    end;
  end;
  try
    Lines.Add(StrNew(StrPCopy(DumChar, LineString)));
    LeCol := TStringColor.Create;
    LeCol.FColor := FCol;
    LeCol.BColor := BCol;
    StringColor.Add(LeCol);
  except
    if PurgeLines <> 0 then
    begin
      for i := 1 to PurgeLines do
      begin
        Lines.Delete(0);
        StringColor.Delete(0);
      end;
      Lines.Pack;
      StringColor.Delete(0);
      try
        Lines.Add(StrNew(StrPCopy(DumChar, LineString)));
        LeCol := TStringColor.Create;
        LeCol.FColor := FCol;
        LeCol.BColor := BCol;
        StringColor.Add(LeCol);
      except
        raise EOutOfResources.Create('Out of Memory at line ' + IntToStr(Lines.Count))
      end;
    end
    else
      raise EOutOfResources.Create('Out of Memory at line ' + IntToStr(Lines.Count))
  end;
  if UpdateDisplay then
  begin
    SetViewportOrg(Canvas.Handle, 0, 0);
    RecalcRange;
    WhereY := Min(Lines.Count - 1, FPageSize);
    Canvas.Font := FFont;
    Canvas.Font.Color := TStringColor(StringColor.Items[Lines.Count -1]).FColor;
    Canvas.Brush.Color := TStringColor(StringColor.Items[Lines.Count -1]).BColor;
    TextOut(Canvas.Handle, 0, FCharSize.Y * WhereY,
            Lines.Items[Lines.Count - 1], StrLen(Lines.Items[Lines.Count -1]));
    ScrollTo(0, FRange.Y);
  end;
end;

procedure TBigText.Delete(Index: longint);
begin
  Lines.Delete(Index);
  StringColor.Delete(Index);
end;

procedure TBigText.Clear;
begin
  Lines.Clear;
  StringColor.Clear;
  RecalcRange;
  Invalidate;
end;

procedure TBigText.Print;
var
  i: integer;
  f: Textfile;
begin
  AssignPrn(f);
  Rewrite(f);
  for i := 0 to (Lines.Count - 1) do
    WriteLn(f, StrPas(Lines.Items[i]));
  System.Close(f);
end;

procedure TBigText.LoadFromFile(FileName: TFileName);
var
  f: TextFile;
  i: integer;
  ReadLine: string;
  DumChar: array[0..255] of char;
  OEMDumChar: array[0..255] of char;
begin
  Clear;
  AssignFile(f, FileName);
  Reset(f);
  while not eof(f) do
  begin
    ReadLn(f, ReadLine);
    while pos(#$9, ReadLine) > 0 do
    begin
      i := pos(#$9, ReadLine);
      System.delete(ReadLine, i, 1);
      while (i mod 8) <> 0 do
      begin
        insert(' ', ReadLine, i);
        inc(i);
      end;
    end;
    StrPCopy(DumChar, ReadLine);
    OEMToAnsi(DumChar, OEMDumChar);
    AddLine(StrPas(OEMDumChar), clWindowText, clWindow, false);
  end;
  CloseFile(f);
  RecalcRange;
  Invalidate;
end;

function TBigText.GetLine(Index: longint): string;
begin
  if Index < Lines.Count then
    GetLine := StrPas(Lines.Items[Index])
  else
    GetLine := '';
end;

procedure TBigText.SetFont(F: TFont);
begin
  FFont.Assign(F);
end;

procedure TBigText.KeyDown(var Key: Word; Shift: TShiftState);
var
  I: Integer;
begin
  inherited KeyDown(Key, Shift);
  if Key <> 0 then
  begin
    for I := 1 to ScrollKeyCount do
      with ScrollKeys[I] do
        if (sKey = Key) and (Ctrl = (Shift = [ssCtrl])) then
        begin
	  DoScroll(SBar, Action, 0);
	  Exit;
        end;
  end;
end;

procedure TBigText.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  SetFocus;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TBigText.WMGetDlgCode(var M: TWMGetDlgCode);
begin
  M.Result := dlgc_WantArrows or dlgc_WantChars;
end;

procedure TBigText.ChangeColor(Index: longint; OldFCol, OldBCol, NewFCol, NewBCol: TColor);
begin
  if (TStringColor(StringColor.Items[Index]).FColor = OldFCol) and
     (TStringColor(StringColor.Items[Index]).BColor = OldBCol) then
  begin
   TStringColor(StringColor.Items[Index]).FColor := NewFCol;
   TStringColor(StringColor.Items[Index]).BColor := NewBCol;
  end;
end;

function TBigText.GetCount: longint;
begin
  if Lines.Count = StringColor.Count then
    GetCount := Lines.Count
  else
    GetCount := -1;
end;

procedure Register;
begin
  RegisterComponents('FreeWare', [TBigText]);
end;

end.
